const int CE = A5, OE = 2, WE = 3,
          CS1 = 4, CS2 = 5,
          DQ0 = 6, DQ1 = 7, DQ2 = 8, DQ3 = 9,
          DQ4 = 10, DQ5 = 11, DQ6 = 12, DQ7 = 13;

const int BUFSIZE = 8;
char inbuffer[BUFSIZE];

const char ERASE[][8] = {
    "W5555AA", "W2AAA55", "W555580", "W5555AA", "W2AAA55"
    };
const char PROGRAM[][8] = {
    "W5555AA", "W2AAA55", "W5555A0"
    };

int fromHex(char c)
{
    if ((c >= 48) && (c <= 57))
        return c - 48;
    if ((c >= 65) && (c <= 70))
        return c - 65 + 10;
    if ((c >= 97) && (c <= 102))
        return c - 97 + 10;
    return 0;
}

void setDataOut()
{
    pinMode(DQ0, OUTPUT);
    pinMode(DQ1, OUTPUT);
    pinMode(DQ2, OUTPUT);
    pinMode(DQ3, OUTPUT);
    pinMode(DQ4, OUTPUT);
    pinMode(DQ5, OUTPUT);
    pinMode(DQ6, OUTPUT);
    pinMode(DQ7, OUTPUT);
}

void setDataIn()
{
    pinMode(DQ0, INPUT);
    pinMode(DQ1, INPUT);
    pinMode(DQ2, INPUT);
    pinMode(DQ3, INPUT);
    pinMode(DQ4, INPUT);
    pinMode(DQ5, INPUT);
    pinMode(DQ6, INPUT);
    pinMode(DQ7, INPUT);
}

int sampleData()
{
    int i = DQ7;
    int data = 0;

    while (1)
    {
        data += digitalRead(i);
        if (i == DQ0)
            break;
        i--;
        data <<= 1;
    }

    return data;
}

int readData()
{
    int data = 0;

    setDataIn();
    delay(1);
    data = sampleData();
    setDataOut();
    return data;
}

void writeData(int data)
{
    int i = DQ0;

    while (1)
    {
        digitalWrite(i, data % 2);
        if (i == DQ7)
            break;
        i++;
        data >>= 1;
    }
}

void setAddress(int high, int low)
{
    writeData(high);
    delayMicroseconds(1);
    digitalWrite(CS2, HIGH);
    delayMicroseconds(1);
    digitalWrite(CS2, LOW);
    writeData(low);
    delayMicroseconds(1);
    digitalWrite(CS1, HIGH);
    delayMicroseconds(1);
    digitalWrite(CS1, LOW);
}

int readOp(int high, int low)
{
    int data;

    digitalWrite(CE, LOW);
    setAddress(high, low);
    delay(1);
    digitalWrite(OE, LOW);
    data = readData();
    delay(1);
    digitalWrite(OE, HIGH);
    digitalWrite(CE, HIGH);

    return data;
}

void writeOp(int high, int low, int data)
{
    digitalWrite(CE, LOW);
    setAddress(high, low);
    delay(1);
    digitalWrite(WE, LOW);
    writeData(data);
    delay(1);
    digitalWrite(WE, HIGH);
    digitalWrite(CE, HIGH);
}

int readCommand(int nread, const char buffer[])
{
    int high, low;

    if (nread < 5)
        return 0;

    high = (fromHex(buffer[1]) << 4) + (fromHex(buffer[2]));
    low = (fromHex(buffer[3]) << 4) + (fromHex(buffer[4]));

    return readOp(high, low);
}

bool writeCommand(int nread, const char buffer[])
{
    int high, low, data;

    if (nread < 7)
        return false;

    high = (fromHex(buffer[1]) << 4) + (fromHex(buffer[2]));
    low = (fromHex(buffer[3]) << 4) + (fromHex(buffer[4]));
    data = (fromHex(buffer[5]) << 4) + (fromHex(buffer[6]));

    writeOp(high, low, data);

    return true;
}

bool waitForCompletion(int high, int low, int written)
{
    int data, then = millis();

    do
    {
        delay(10);
        data = readOp(high, low);
    } while (
        ((data & 0x80) != (written & 0x80)) &&
        ((data & 0x20) == 0) &&
        (millis() - then < 2000)
        );

    return readOp(high, low) == written;
}

bool chipErase()
{
    for (int i = 0; i < 5; i++)
    {
        writeCommand(7, ERASE[i]);
    }
    writeOp(0x55, 0x55, 0x30);
    return waitForCompletion(0, 0, 0xff);
}

bool sectorErase(int nread, const char buffer[])
{
    int high;

    if (nread < 2)
        return false;

    // 3-bit number shifted to A16...A14 with A16 discarded.
    // Thus, only sectors from 0 to 3 are supported.

    high = (fromHex(buffer[1]) << 6) & 0xff;

    for (int i = 0; i < 5; i++)
    {
        writeCommand(7, ERASE[i]);
    }
    writeOp(high, 0, 0x30);

    // Result of erase if 0xff written to all locations.

    return waitForCompletion(high, 0, 0xff);
}

bool program(int nread, const char buffer[])
{
    int high, low, data;

    if (nread < 7)
        return false;

    high = (fromHex(buffer[1]) << 4) + (fromHex(buffer[2]));
    low = (fromHex(buffer[3]) << 4) + (fromHex(buffer[4]));
    data = (fromHex(buffer[5]) << 4) + (fromHex(buffer[6]));

    for (int i = 0; i < 3; i++)
    {
        writeCommand(7, PROGRAM[i]);
    }
    writeOp(high, low, data);

    return waitForCompletion(high, low, data);
}

void setup()
{
    Serial.begin(115200);
    pinMode(CE, OUTPUT);
    pinMode(OE, OUTPUT);
    pinMode(WE, OUTPUT);
    pinMode(CS1, OUTPUT);
    pinMode(CS2, OUTPUT);
    setDataOut();

    // Initial state for the flash device.

    digitalWrite(CE, HIGH);
    digitalWrite(OE, HIGH);
    digitalWrite(WE, HIGH);

    // Initial state for the latches.

    digitalWrite(CS1, LOW);
    digitalWrite(CS2, LOW);

    // Interface loop.

    Serial.println("?");
    Serial.setTimeout(3600000); // 1 hour
}

void loop()
{
    int nread = 0;

    if (nread = Serial.readBytesUntil('\n', inbuffer, BUFSIZE))
    {
        switch (inbuffer[0])
        {
            case 'R':
            Serial.println(readCommand(nread, inbuffer), HEX);
            break;

            case 'W':
            if (writeCommand(nread, inbuffer))
                Serial.println("W");
            else
                Serial.println("C");
            break;

            case 'E':
            if (chipErase())
                Serial.println("E");
            else
                Serial.println("C");
            break;

            case 'S':
            if (sectorErase(nread, inbuffer))
                Serial.println("S");
            else
                Serial.println("C");
            break;

            case 'P':
            if (program(nread, inbuffer))
                Serial.println("P");
            else
                Serial.println("C");
            break;

            default:
            Serial.println("?");
            break;
        }
    }
}

extern "C" void __cxa_pure_virtual(void) {
    while(1);
}

// tabstop=4 expandtab shiftwidth=4
